#include <algorithm>
#include "gbpmm.h"
#include "guessgen.h"
#include "timer.h"

using namespace std;

// Define here a constructor for fixed questions.
template<int colors>
Configuration<colors>::Configuration(int j) {
  if (j >= 0 && j < colors)
    for (int i = 0; i < PEGS; ++i)
      pegs[i] = first_peg_color + j;
}

ostream& operator<<(ostream &os, const DQuestion &q) {
  os << q.first << ' ' << q.second[0] << q.second[0] << q.second[0]
     << "..." << q.second[0];
  return os;
}

long power(long base, long exp) {
  int result = 1;
  while (exp) {
    if (exp & 1)
      result *= base;
    exp >>= 1;
    base *= base;
  }
  return result;
}

// Masets are sorted in the decreasing order of they size.
static bool compareMaSets(pair<MaSet*, DMaAnswer> p1,
                          pair<MaSet*, DMaAnswer> p2) {
  return p1.first->size() > p2.first->size();
}

class DAnsweredMaSet {
  private:
    vector<pair<MaSet*, DMaAnswer> > masets;
    DQuestion asked_question;
    void init() {
      masets.resize(ANSWERS);
      for (MaSetSize i = 0; i < ANSWERS; ++i) {
        masets[i].first = new MaSet;
        masets[i].second = DMaAnswer(i);
      }
    }
  public:
    DAnsweredMaSet(const MaSet &in_maset, const DQuestion &guess) {
      init();
      asked_question = guess;
      for (MaSetSize i = 0; i < in_maset.size(); ++i)
        masets[DMaAnswer(in_maset[i], guess).code()].first
          ->push_back(in_maset[i]);
      sort(masets.begin(), masets.end(), compareMaSets);
    }
    DAnsweredMaSet(const DQuestion &guess) {
      init();
      asked_question = guess;
      Secret conf(0), first_conf(0);
      do
        masets[DMaAnswer(conf, guess).code()].first->push_back(conf);
      while (++conf != first_conf);
      sort(masets.begin(), masets.end(), compareMaSets);
    }
    ~DAnsweredMaSet() {
      for (MaSetSize i = 0; i < masets.size(); ++i)
        delete masets[i].first;
    }
    MaSet& getMaSet(int i) {
      return *masets[i].first;
    }
    DMaAnswer& getMaAnswer(int i) {
      return masets[i].second;
    }
    MaSetSize size() const {
      return masets.size();
    }
    MaSetSize max_size() const {
      return masets[0].first->size();
    }
    DQuestion& getAskedQuestion() {
      return asked_question;
    }
    friend bool
    compareDAnsweredMaSets(const DAnsweredMaSet*, const DAnsweredMaSet*);
};

// Answered masets are sorted in the increasing order of the size of the
// biggest maset.
bool compareDAnsweredMaSets(const DAnsweredMaSet *p1,
                            const DAnsweredMaSet *p2) {
  return p1->max_size() < p2->max_size();
}

class DAllAnswers {
  private:
    vector<DAnsweredMaSet*> answered_masets;
  public:
    DAllAnswers(const MaSet *maset, GuessGenerator *gg, MaSetSize upper_size) {
      Question const *guess;
      DAnsweredMaSet *ans;
      gg->reset();
      while ((guess = gg->get()) != 0) {
        for (int i = 0; i < FIXED_QUESTIONS; ++i) {
          Question q(i);
          if (q != *guess) {
            if (maset)
              ans = new DAnsweredMaSet(*maset, make_pair(*guess, q));
            else
              ans = new DAnsweredMaSet(make_pair(*guess, q));
            // Do not store answers for hopeless questions.
            if (ans->max_size() > upper_size)
              delete ans;
            //#define K_SEARCH
            #ifdef K_SEARCH
            else if (ans->max_size() < upper_size) {
              upper_size = ans->max_size();
              answered_masets.clear();
              answered_masets.push_back(ans);
            }
            #endif
            else
              answered_masets.push_back(ans);
            #ifndef K_SEARCH
            sort(answered_masets.begin(), answered_masets.end(),
                 compareDAnsweredMaSets);
            #endif
          }
        }
      }
    }
    ~DAllAnswers() {
      for (MaSetSize i = 0; i < answered_masets.size(); ++i)
        delete answered_masets[i];
    }
    DAnsweredMaSet& getAnsweredMaSet(int i) {
      return *answered_masets[i];
    }
    MaSetSize size() const {
      return answered_masets.size();
    }
};

class Bounds {
  private:
    MaSetSize lower_bound_table[QUESTIONS + 1];
  public:
    Bounds() {
      lower_bound_table[0] = 0;
      for (int i = 0; i < QUESTIONS; ++i)
        lower_bound_table[i + 1] = (ANSWERS - 1) * lower_bound_table[i] + 1;
    }
    // The information-theoretic lower bound. The maximal size of the maset,
    // which could be solved in guess_number questions.
    MaSetSize lower(int guess_number) const {
      return lower_bound_table[guess_number];
    }
    // The upper bound is the maximal size of the maset, for which we can
    // prove that it is solvable in guess_number questions.
    MaSetSize upper(int guess_number) const {
      return guess_number;
    }
};

const static Bounds bound;

class DGuessGenWrapper {
  public:
    GuessGenerator *gg;
    DGuessGenWrapper(const DQSeq &asked) {
      if (asked.size() == 0)
        gg = new BprimAlmostStaticGuessGenerator(0);
      else if (asked.size() == 1)
        gg = new BprimAlmostStaticGuessGenerator(1);
      else
        gg = new ExhaustiveGuessGenerator();
    }
    ~DGuessGenWrapper() {
      delete gg;
    }
};

static bool Dsolvable(const MaSet *maset, GuessGenerator *gg,
                      DQSeq &asked, vector<DMaAnswer> &mans,
                      int guesses_number) {
  guesses_number--;
  DAllAnswers ans(maset, gg, bound.lower(guesses_number));
  for (MaSetSize i = 0; i < ans.size(); ++i) {
    asked.push_back(ans.getAnsweredMaSet(i).getAskedQuestion());
    DGuessGenWrapper ggw(asked);
    for (MaSetSize j = 0; j < ans.getAnsweredMaSet(i).size(); ++j) {
      if (ans.getAnsweredMaSet(i).getMaSet(j).size() <=
          bound.upper(guesses_number)) {
        #ifdef PRINT_STRATEGY // BEGIN strategy printing
        cout << "----------------------------------------\n";
        if (asked.size() > 1) {
          cout << "Former questions and answers\n";
          for (MaSetSize k = 0; k < asked.size() - 1; ++k)
            cout << asked[k] << ' ' << mans[k] << '\n';
          cout << '\n';
        }
        cout << "Present question " << asked.back() << '\n';
        for (MaSetSize k = 0; k < ans.getAnsweredMaSet(i).size(); ++k)
          if (ans.getAnsweredMaSet(i).getMaSet(k).size())
            cout << "\nAnswer "
                 << ans.getAnsweredMaSet(i).getMaAnswer(k)
                 << ", possible first " << PEGS << " pegs\n"
                 << ans.getAnsweredMaSet(i).getMaSet(k);
          // else
          //  cout << "\nThe answer "
          //       << ans.getAnsweredMaSet(i).getMaAnswer(k)
          //       << " gives the empty maset\n";*/
        #endif // END strategy printing
        asked.pop_back();
        // Answered masets are sorted in the increasing order of the size of
        // the biggest maset.  Masets inside answered maset are sorted in the
        // decreasing order of they size.  If the size of the considered maset
        // is less or equal to the upper bound, then the size of all subsequent
        // masets are also less or equal.
        return true;
      }
      mans.push_back(ans.getAnsweredMaSet(i).getMaAnswer(j));
      bool solution_found = Dsolvable(&ans.getAnsweredMaSet(i).getMaSet(j),
                                      ggw.gg, asked, mans, guesses_number);
      mans.pop_back();
      // If not solvable, we consider the next question.
      if (!solution_found)
        break;
    }
    asked.pop_back();
  }
  return false;
}

static char* plural(unsigned i) {
  static char s0[1] ="";
  static char s1[2] = "s";
  return i == 1 ? s0 : s1;
}

int main() {
  Timer t;

  #if FUNCTION == DIFFERENTIAL_FUNCTION
  const char *G = "b'";
  cout << PEGS << " peg" << plural(PEGS) << "\n"
       << COLORS << " color" << plural(COLORS) << "\n"
       << QUESTION_COLORS << " color" << plural(QUESTION_COLORS) << " in questions\n"
       << QUESTIONS-1 << " question" << plural(QUESTIONS-1) << '\n'
       << FIXED_QUESTIONS << " fixed question" << plural(FIXED_QUESTIONS) << "\n";
  for (int i = 0; i < FIXED_QUESTIONS; ++i)
    cout << "\t" << Question(i) << Question(i) << "...\n";
  cout << "computing " << G << endl;

  const int qa = QUESTIONS;

  bool result;
  DQSeq already_asked;
  already_asked.reserve(qa);

  MaSetSize conf_number = power(COLORS, PEGS);
  if (conf_number <= bound.upper(qa))
    result = true;
  else if (conf_number > bound.lower(qa))
    result = false;
  else {
    DGuessGenWrapper gw(already_asked);
    vector<DMaAnswer> ma;
    result = Dsolvable(0, gw.gg, already_asked, ma, qa);
  }

  #ifdef PRINT_STRATEGY
  cout << "----------------------------------------\n";
  #endif
  cout << "result " << G << "(" << PEGS << "," << MA_COLORS << ") "
       << (result ? "<= " : "> ") << QUESTIONS-1 << '\n'
       << "computation time " << t.getTime() << endl;
  #endif
  return 0;
}
