#include <algorithm>
#include "gbpmm.h"
#include "guessgen.h"
#include "timer.h"

using namespace std;

// Define here a constructor for fixed questions.
template<int colors>
Configuration<colors>::Configuration(int j) {
  // Fixed questions for functions $b$ and $b^*$
  if (j >= 0 && j < colors)
    for (int i = 0; i < PEGS; ++i)
      pegs[i] = first_peg_color + j;
}

long power(long base, long exp) {
  long result = 1;
  while (exp) {
    if (exp & 1)
      result *= base;
    exp >>= 1;
    base *= base;
  }
  return result;
}

// Masets are sorted in the decreasing order of they size.
static bool compareMaSets(pair<MaSet*, MaAnswer> p1,
                          pair<MaSet*, MaAnswer> p2) {
  return p1.first->size() > p2.first->size();
}

class AnsweredMaSet {
  private:
    vector<pair<MaSet*, MaAnswer> > masets;
    Question asked_question;
    void init() {
      masets.resize(ANSWERS);
      for (MaSetSize i = 0; i < ANSWERS; ++i) {
        masets[i].first = new MaSet;
        masets[i].second = MaAnswer(i);
      }
    }
  public:
    AnsweredMaSet(const MaSet &in_maset, const Question &guess) {
      init();
      asked_question = guess;
      for (MaSetSize i = 0; i < in_maset.size(); ++i)
        masets[MaAnswer(in_maset[i], guess).code()].first
          ->push_back(in_maset[i]);
      sort(masets.begin(), masets.end(), compareMaSets);
    }
    AnsweredMaSet(const Question &guess) {
      init();
      asked_question = guess;
      Secret conf(0), first_conf(0);
      do
        masets[MaAnswer(conf, guess).code()].first->push_back(conf);
      while (++conf != first_conf);
      sort(masets.begin(), masets.end(), compareMaSets);
    }
    ~AnsweredMaSet() {
      for (MaSetSize i = 0; i < masets.size(); ++i)
        delete masets[i].first;
    }
    MaSet& getMaSet(int i) {
      return *masets[i].first;
    }
    MaAnswer& getMaAnswer(int i) {
      return masets[i].second;
    }
    MaSetSize size() const {
      return masets.size();
    }
    MaSetSize max_size() const {
      return masets[0].first->size();
    }
    Question& getAskedQuestion() {
      return asked_question;
    }
    friend bool
    compareAnsweredMaSets(const AnsweredMaSet*, const AnsweredMaSet*);
};

// Answered masets are sorted in the increasing order of the size of the
// biggest maset.
bool compareAnsweredMaSets(const AnsweredMaSet *p1, const AnsweredMaSet *p2) {
  return p1->max_size() < p2->max_size();
}

class AllAnswers {
  private:
    vector<AnsweredMaSet*> answered_masets;
  public:
    AllAnswers(const MaSet *maset, GuessGenerator *gg, MaSetSize upper_size) {
      Question const *guess;
      AnsweredMaSet *ans;
      gg->reset();
      while ((guess = gg->get()) != 0) {
        if (maset)
          ans = new AnsweredMaSet(*maset, *guess);
        else
          ans = new AnsweredMaSet(*guess);
        // Do not store answers for hopeless questions.
        if (ans->max_size() > upper_size)
          delete ans;
        #if FUNCTION == KNUTH_GREEDY_STRATEGY_FUNCTION
        else if (ans->max_size() < upper_size) {
          upper_size = ans->max_size();
          answered_masets.clear();
          answered_masets.push_back(ans);
        }
        #endif
        else
          // Now ans->max_size() == upper_size for the Knuth-Greedy strategy
          // and ans->max_size() <= upper_size in the other case.
          answered_masets.push_back(ans);
      }
      #if FUNCTION != KNUTH_GREEDY_STRATEGY_FUNCTION
      sort(answered_masets.begin(), answered_masets.end(),
           compareAnsweredMaSets);
      #endif
    }
    ~AllAnswers() {
      for (MaSetSize i = 0; i < answered_masets.size(); ++i)
        delete answered_masets[i];
    }
    AnsweredMaSet& getAnsweredMaSet(int i) {
      return *answered_masets[i];
    }
    MaSetSize size() const {
      return answered_masets.size();
    }
};

class Bounds {
  private:
    MaSetSize lower_bound_table[QUESTIONS + 1];
  public:
    Bounds() {
      lower_bound_table[0] = 0;
      for (int i = 0; i < QUESTIONS; ++i)
        lower_bound_table[i + 1] = (ANSWERS - 1) * lower_bound_table[i] + 1;
    }
    // The information-theoretic lower bound. The maximal size of the maset,
    // which could be solved in guess_number questions.
    MaSetSize lower(int guess_number) const {
      return lower_bound_table[guess_number];
    }
    // The upper bound is the maximal size of the maset, for which we can
    // prove that it is solvable in guess_number questions.
    MaSetSize upper(int guess_number) const {
      return guess_number;
    }
};

const static Bounds bound;

class GuessGenWrapper {
  public:
    GuessGenerator *gg;
    GuessGenWrapper(const QSeq &asked) {
      if (FUNCTION == ONE_ADDITIONAL_COLOR_FUNCTION && asked.size() == 0)
        gg = new Lemma2bGuessGenerator();
      else if (asked.size() < FIXED_QUESTIONS)
        gg = new FixedGuessGenerator(asked.size());
      else if (asked.size() < FIXED_QUESTIONS + ITESTS)
        gg = new NautyGuessGenerator(asked);
      else
        gg = new ExhaustiveGuessGenerator();
    }
    ~GuessGenWrapper() {
      delete gg;
    }
};

static bool solvable(const MaSet *maset, GuessGenerator *gg,
                     QSeq &asked, vector<MaAnswer> &mans,
                     int guesses_number) {
  guesses_number--;
  AllAnswers ans(maset, gg, bound.lower(guesses_number));
  for (MaSetSize i = 0; i < ans.size(); ++i) {
    asked.push_back(ans.getAnsweredMaSet(i).getAskedQuestion());
    GuessGenWrapper ggw(asked);
    for (MaSetSize j = 0; j < ans.getAnsweredMaSet(i).size(); ++j) {
      if (ans.getAnsweredMaSet(i).getMaSet(j).size() <=
          bound.upper(guesses_number)) {
        #ifdef PRINT_STRATEGY // BEGIN strategy printing
        cout << "----------------------------------------\n";
        if (asked.size() > 1) {
          cout << "Former questions and answers\n";
          for (MaSetSize k = 0; k < asked.size() - 1; ++k)
            cout << asked[k] << ' ' << mans[k] << '\n';
          cout << '\n';
        }
        cout << "Present question " << asked.back() << '\n';
        for (MaSetSize k = 0; k < ans.getAnsweredMaSet(i).size(); ++k)
          if (ans.getAnsweredMaSet(i).getMaSet(k).size())
            cout << "\nThe answer "
                 << ans.getAnsweredMaSet(i).getMaAnswer(k)
                 << " gives the maset\n"
                 << ans.getAnsweredMaSet(i).getMaSet(k);
          else
            cout << "\nThe answer "
                 << ans.getAnsweredMaSet(i).getMaAnswer(k)
                 << " gives the empty maset\n";
        #endif // END strategy printing
        asked.pop_back();
        // Answered masets are sorted in the increasing order of the size of
        // the biggest maset.  Masets inside answered maset are sorted in the
        // decreasing order of they size.  If the size of the considered maset
        // is less or equal to the upper bound, then the size of all subsequent
        // masets are also less or equal.
        return true;
      }
      mans.push_back(ans.getAnsweredMaSet(i).getMaAnswer(j));
      bool solution_found = solvable(&ans.getAnsweredMaSet(i).getMaSet(j),
                                     ggw.gg, asked, mans, guesses_number);
      mans.pop_back();
      // If not solvable, we consider the next question.
      if (!solution_found)
        break;
    }
    asked.pop_back();
  }
  return false;
}

char* plural(unsigned i) {
  static char s0[1] ="";
  static char s1[2] = "s";
  return i == 1 ? s0 : s1;
}

int main() {
  Timer t;

  #if FUNCTION == NORMAL_FUNCTION
  const char *G = "the exact value or the upper bound";
  #elif FUNCTION == ONE_ADDITIONAL_COLOR_FUNCTION
  const char *G = "the lower bound";
  #elif FUNCTION == KNUTH_GREEDY_STRATEGY_FUNCTION
  const char *G = "the upper bound using the Knuth-Greedy strategy";
  #endif
  cout << PEGS << " peg" << plural(PEGS) << "\n"
       << COLORS << " color" << plural(COLORS) << "\n"
       << QUESTION_COLORS << " color" << plural(QUESTION_COLORS) << " in questions\n"
       << QUESTIONS << " question" << plural(QUESTIONS) << '\n'
       << FIXED_QUESTIONS << " fixed question" << plural(FIXED_QUESTIONS) << "\n";
  #if FUNCTION != ONE_ADDITIONAL_COLOR_FUNCTION
  for (int i = 0; i < FIXED_QUESTIONS; ++i)
    cout << "\t" << Question(i) << '\n';
  #endif
  cout << ITESTS << " question" << plural(ITESTS) << " tested for isomorphism\n";
  cout << "computing " << G << endl;

  const int qa = QUESTIONS;

  bool result;
  QSeq already_asked;
  already_asked.reserve(qa);

  MaSetSize conf_number = power(COLORS, PEGS);
  if (conf_number <= bound.upper(qa))
    result = true;
  else if (conf_number > bound.lower(qa))
    result = false;
  else {
    GuessGenWrapper gw(already_asked);
    vector<MaAnswer> ma;
    result = solvable(0, gw.gg, already_asked, ma, qa);
  }

  #ifdef PRINT_STRATEGY
  cout << "----------------------------------------\n";
  #endif
  cout << (result ? "a strategy FOUND\n" : "a strategy NOT FOUND\n")
       << "computation time " << t.getTime() << endl;
  return 0;
}
