#ifndef _MASTERMIND_H_
#define _MASTERMIND_H_

#include <iostream>
#include <vector>

#define NORMAL_FUNCTION                 0
#define ONE_ADDITIONAL_COLOR_FUNCTION   1
#define KNUTH_GREEDY_STRATEGY_FUNCTION  3
#define DIFFERENTIAL_FUNCTION           9

#ifndef ITESTS
  #define ITESTS 0
#endif

#ifndef FIXED_QUESTIONS
  #define FIXED_QUESTIONS 0
#endif

#if FUNCTION == DIFFERENTIAL_FUNCTION
  #define ANSWERS (2*PEGS+1) // the number of possible answers
#else
  #define ANSWERS (PEGS+1) // the number of possible answers
#endif

#if FUNCTION == ONE_ADDITIONAL_COLOR_FUNCTION
  #define COLORS          (MA_COLORS)
  #define QUESTION_COLORS (MA_COLORS+1)
#else
  #define COLORS          (MA_COLORS)
  #define QUESTION_COLORS (MA_COLORS)
#endif

// Pegs colors are 'x', 'y', 'z', etc.
typedef char Peg;

template<int colors> class Configuration {
  private:
    static const Peg first_peg_color = colors <= 3 ? 'x' : 'z' - colors + 1;
    static const Peg last_peg_color = first_peg_color + colors - 1;
    Peg pegs[PEGS];
  public:
    // Constructor for the first question
    Configuration() {
      for (int i = 0; i < PEGS; ++i)
        pegs[i] = first_peg_color;
    }
    // Constructor for fixed questions, it must be defined elsewhere.
    Configuration(int j);
    // Constructor in Lemma2bGuessGenerator
    Configuration(const Peg s[PEGS]) {
      for (int i = 0; i < PEGS; ++i)
        pegs[i] = s[i];
    }
    Configuration& operator++() {
      for(int i = PEGS - 1; i >= 0; --i) {
        if (pegs[i] == last_peg_color)
          pegs[i] = first_peg_color;
        else {
          ++pegs[i];
          break;
        }
      }
      return *this;
    }
    bool operator!=(const Configuration& conf) {
      for (int i = 0; i < PEGS; ++i)
        if (pegs[i] != conf.pegs[i])
          return true;
      return false;
    }
    Peg operator[](int i) const {
      return pegs[i];
    }
    static Peg firstPegColor() {
      return first_peg_color;
    }
    static Peg lastPegColor() {
      return last_peg_color;
    }
    template<int c0> friend ::std::ostream&
    operator<<(::std::ostream&, const Configuration<c0>&);
    template<int c0, int c1> friend int
    answer(const Configuration<c0>&, const Configuration<c1>&);
};

template<int colors> ::std::ostream&
operator<<(::std::ostream& os, const Configuration<colors>& conf) {
  for (int i = 0; i < PEGS; ++i)
    os << conf.pegs[i];
  return os;
}

template<int conf_colors, int guess_colors> class Answer {
  private:
    int blacks; // the number of black pegs
  public:
    Answer() {
    }
    Answer(int b) : blacks(b) {
    }
    Answer(const Configuration<conf_colors> &conf,
           const Configuration<guess_colors> &guess) {
      blacks = 0;
      for (int i = 0; i < PEGS; ++i)
        if (conf[i] == guess[i])
          ++blacks;
    }
    int code() {
      return blacks;
    }
    template<int conf_c, int guess_c> friend ::std::ostream&
    operator<<(::std::ostream&, const Answer<conf_c, guess_c>&);
};

template<int conf_c, int guess_c> ::std::ostream&
operator<<(::std::ostream &os, const Answer<conf_c, guess_c> &ans) {
  os << ans.blacks;
  return os;
}

template<int conf_colors, int guess_colors> class DAnswer {
  private:
    int difference; // the number of black pegs
  public:
    DAnswer() {
    }
    DAnswer(int d) : difference(d - PEGS) {
    }
    DAnswer(const Configuration<conf_colors> &conf,
            const ::std::pair<Configuration<guess_colors>,
                              Configuration<guess_colors> > &guess) {
      difference = 0;
      for (int i = 0; i < PEGS; ++i)
        difference += (conf[i] == guess.first[i]) -
                      (conf[i] == guess.second[i]);
    }
    int code() {
      return difference + PEGS;
    }
    template<int conf_c, int guess_c> friend ::std::ostream&
    operator<<(::std::ostream&, const DAnswer<conf_c, guess_c>&);
};

template<int conf_c, int guess_c> ::std::ostream&
operator<<(::std::ostream &os, const DAnswer<conf_c, guess_c> &ans) {
  os << ans.difference;
  return os;
}

typedef Configuration<COLORS> Secret;
typedef Configuration<QUESTION_COLORS> Question;
typedef ::std::pair<Question, Question> DQuestion;
typedef ::std::vector<Secret> MaSet;
typedef ::std::vector<Question> QSeq;
typedef ::std::vector<DQuestion> DQSeq;
typedef MaSet::size_type MaSetSize;
typedef QSeq::size_type QSeqSize;
typedef DQSeq::size_type DQSeqSize;
typedef Answer<COLORS, QUESTION_COLORS> MaAnswer;
typedef DAnswer<COLORS, QUESTION_COLORS> DMaAnswer;

template<int colors> ::std::ostream&
operator<<(::std::ostream &os, const ::std::vector<Configuration<colors> > &maset) {
  for (MaSetSize i = 0; i < maset.size(); ++i)
    os << maset[i] << '\n';
  os << ::std::flush;
  return os;
}

long power(long, long);

#endif
