#ifndef _GUESSGEN_H_
#define _GUESSGEN_H_

#include <map>
#include "gbpmm.h"

#define MAXN (QUESTION_COLORS + PEGS * (FIXED_QUESTIONS + ITESTS + 1))
#include "nauty.h"

// Wrapper for nauty graph
class Graph {
  protected:
    int n, m;
    graph g[MAXN*MAXM];
  public:
    int getVerticesNumber() {
      return n;
    }
    graph* getGraph() {
      return g;
    }
    Graph(int vertices_number) {
      n = vertices_number;
      m = (n + WORDSIZE - 1) / WORDSIZE;
    }
    bool operator==(const Graph& s) {
      for (int i = 0; i < n * m; ++i)
        if (g[i] != s.g[i])
          return false;
      return true;
    }
    bool operator<(const Graph& s) const {
      for (int i = 0; i < n * m; ++i)
        if (g[i] < s.g[i])
          return true;
        else if (g[i] > s.g[i])
          return false;
      return false;
    }
};

// Graph representing a question sequence
class QGraph : public Graph {
  private:
    // tables used by nauty
    set *gv[MAXN];
    int lab[MAXN], ptn[MAXN], orbits[MAXN];

    void addHypherEdge(int x, int y, int z) {
      // A triple (x, y, z) represents a peg,
      // x is the column number, y is the row number and z is its color.
      // Add two edges {x+QUESTION_COLORS+PEGS+y*PEGS, x+QUESTION_COLORS},
      //               {x+QUESTION_COLORS+PEGS+y*PEGS, z}.
      ADDELEMENT(gv[x+QUESTION_COLORS+PEGS+y*PEGS], x+QUESTION_COLORS);
      ADDELEMENT(gv[x+QUESTION_COLORS+PEGS+y*PEGS], z);
      ADDELEMENT(gv[x+QUESTION_COLORS], x+QUESTION_COLORS+PEGS+y*PEGS);
      ADDELEMENT(gv[z],                 x+QUESTION_COLORS+PEGS+y*PEGS);
    }
    void delHypherEdge(int x, int y, int z) {
      // A triple (x, y, z) represents a peg,
      // x is the column number, y is the row number and z is its color.
      // Delete two edges {x+QUESTION_COLORS+PEGS+y*PEGS, x+QUESTION_COLORS},
      //                  {x+QUESTION_COLORS+PEGS+y*PEGS, z}.
      DELELEMENT(gv[x+QUESTION_COLORS+PEGS+y*PEGS], x+QUESTION_COLORS);
      DELELEMENT(gv[x+QUESTION_COLORS+PEGS+y*PEGS], z);
      DELELEMENT(gv[x+QUESTION_COLORS], x+QUESTION_COLORS+PEGS+y*PEGS);
      DELELEMENT(gv[z],                 x+QUESTION_COLORS+PEGS+y*PEGS);
    }
    void initLabPtn() {
      // Initialize tables used by nauty.
      for (int i = 0; i < n; ++i) {
        lab[i] = i;
        ptn[i] = 1;
      }
      #if COLORS != QUESTION_COLORS
      ptn[COLORS - 1] = 0; // Do not map additional color to any other color.
      #endif
      for (int i = QUESTION_COLORS - 1; i < n; i += PEGS)
        ptn[i] = 0;
    }
  public:
    QGraph(int rows) : Graph(QUESTION_COLORS + PEGS + PEGS * rows) {
      // nauty_check(WORDSIZE, m, n, NAUTYVERSIONID);
      for (int i = 0; i < n; ++i) {
        gv[i] = GRAPHROW(g, i, m);
        EMPTYSET(gv[i], m);
      }
    }
    void addQuestion(int row, const Question &q) {
       for (int i = 0; i < PEGS; ++i)
         addHypherEdge(i, row, q[i] - q.firstPegColor());
    }
    void delQuestion(int row, const Question &q) {
       for (int i = 0; i < PEGS; ++i)
         delHypherEdge(i, row, q[i] - q.firstPegColor());
    }
    void callNauty(Graph *gc) {
      #define WRKSPC (50 * MAXM)
      static DEFAULTOPTIONS_GRAPH(options);
      static setword workspace[WRKSPC];
      static statsblk stats;
      options.getcanon = TRUE;
      options.defaultptn = FALSE;
      initLabPtn();
      nauty(g, lab, ptn, NULL, orbits, &options, &stats,
            workspace, WRKSPC, m, n, gc->getGraph());
      #undef WRKSPC
    }
};

typedef ::std::map<Graph, Question> QSet;
typedef ::std::pair<Graph, Question> QPair;

class GuessGenerator {
  public:
    virtual Question const * get() = 0;
    virtual void reset() = 0;
    virtual ~GuessGenerator() {}
};

class ExhaustiveGuessGenerator : public GuessGenerator {
  private:
    const Question first_q;
    Question current_q, q;
    int qn, qa;
  public:
    ExhaustiveGuessGenerator() : first_q() {
      qn = power(QUESTION_COLORS, PEGS);
    }
    void reset() {
      qa = qn;
      current_q = first_q;
    }
    Question const * get() {
      q = current_q;
      ++current_q;
      --qa;
      if (qa >= 0)
        return &q;
      else
        return 0;
    }
};

class FixedGuessGenerator : public GuessGenerator {
  private:
    const Question q;
    bool done;
  public:
    FixedGuessGenerator(int r) : q(r), done(false) {
    }
    void reset() {
      done = false;
    }
    Question const * get() {
      if (done)
        return 0;
      else {
        done = true;
        return &q;
      }
    }
};

class BprimAlmostStaticGuessGenerator : public GuessGenerator {
  private:
    Question q;
    Peg s[PEGS];
    bool done;
  public:
    BprimAlmostStaticGuessGenerator(int r) : done(false) {
      if (r == 0) {
        for (int i = 0; i < PEGS; ++i)
          s[i] = Question::firstPegColor();
      }
      else if (r == 1) {
        for (int i; i < PEGS/2; ++i)
          s[i] = Question::firstPegColor();
        for (int i = PEGS/2; i < PEGS; ++i)
          s[i] = Question::firstPegColor() + 1;
      }
      q = Question(s);
    }
    void reset() {
      done = false;
    }
    Question const * get() {
      if (done)
        return 0;
      else {
        done = true;
        return &q;
      }
    }
};

class Lemma2bGuessGenerator : public GuessGenerator {
  private:
    Question q;
    int idx;
    Peg s[PEGS];
  public:
    Lemma2bGuessGenerator() : idx(0) {
      for (int i = 0; i < PEGS; ++i)
        s[i] = Question::firstPegColor();
    }
    void reset() {
      idx = 0;
    }
    Question const * get() {
      if (idx < PEGS + 1) {
        q = Question(s);
        ++idx;
        s[PEGS - idx] = Question::lastPegColor();
        return &q;
      }
      else
        return 0;
    }
};

class NautyGuessGenerator : public GuessGenerator {
  private:
    QSet question_set;
    int level;
    QGraph q_graph;
    Graph canon_graph;
    const Question first_question;
    QSet::iterator it;
  public:
    NautyGuessGenerator(const QSeq &asked) :
      level(asked.size()),
      q_graph(level + 1),
      canon_graph(q_graph.getVerticesNumber()),
      first_question() {
      // Initialize graph which already asked questions.
      for (int i = 0; i < level; ++i)
        q_graph.addQuestion(i, asked[i]);
      // Compute all non-isomorphic questions extending asked questions.
      Question question = first_question;
      do {
        q_graph.addQuestion(level, question);
        q_graph.callNauty(&canon_graph);
        question_set.insert(QPair(canon_graph, question));
        q_graph.delQuestion(level, question);
      } while (++question != first_question);
      // Do not repeat the same questions.
      for (int i = 0; i < level; ++i) {
        q_graph.addQuestion(level, asked[i]);
        q_graph.callNauty(&canon_graph);
        question_set.erase(canon_graph);
        q_graph.delQuestion(level, asked[i]);
      }
    }
    void reset() {
      // Set iterator to the first question.
      it = question_set.begin();
    }
    Question const * get() {
      if (it != question_set.end()) {
        Question *q = &(*it).second;
        ++it;
        return q;
      }
      else
        return 0;
    }
};

#endif
